﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace ReadyQueueDemo
{
	public partial class ConcurrentList : UserControl
	{
		IntPtr _Handle = IntPtr.Zero;
		Size _Size = Size.Empty;
		readonly Font _Font = SystemFonts.DialogFont;
		readonly SynchronizationContext _SyncContext = null;
		volatile int _VerticalScrollPosition = 0;
		volatile int _ViewportHeight = 0;

		public ConcurrentList()
		{
			InitializeComponent();

			_SyncContext = SynchronizationContext.Current ?? new SynchronizationContext();

			BackColor = SystemColors.Window;
			DoubleBuffered = true;
			AutoScroll = true;
			VScroll = true;

			Load += ConcurrentList_Load;
			SizeChanged += ConcurrentList_SizeChanged;
			Scroll += ConcurrentList_Scroll;

			SetVerticalScroll();
		}

		void ConcurrentList_Load( object sender, EventArgs e )
		{
			_Handle = Handle;
		}

		void ConcurrentList_SizeChanged( object sender, EventArgs e )
		{
			_Size = Size;
			Invalidate();
		}

		void ConcurrentList_Scroll( object sender, ScrollEventArgs e )
		{
			if ( e.ScrollOrientation != ScrollOrientation.VerticalScroll ) return;

			_VerticalScrollPosition = e.NewValue;

			Invalidate();
		}

		void SetVerticalScroll()
		{
			_SyncContext.Post( o => { AutoScrollMinSize = new Size( 0, _ViewportHeight ); }, null );
		}

		public void Clear()
		{
			lock ( key )
			{
				_EntryList.Clear();
				_ThreadDatas.Clear();
			}

			ConcurrentPaint();

			SetVerticalScroll();
		}

		//-----------------------------------------------------------------------------------------
		// Threads

		class ThreadData
		{
			public int MTId { get; protected set; }
			public int Index { get; protected set; }
			public string Name { get; protected set; }
			public float ThreadX { get; set; }
			public float TextX { get; set; }

			public ThreadData( int index, string name )
			{
				MTId = Thread.CurrentThread.ManagedThreadId;
				Index = index;
				Name = name;
			}

			public ThreadData( string name )
				: this( -1, name )
			{
			}
		}

		SortedList<int, ThreadData> _ThreadDatas = new SortedList<int, ThreadData>();
		object key = new object();

		internal void RegisterThread( int index, string name )
		{
			var td = new ThreadData( index, name );
			lock ( key ) _ThreadDatas.Add( td.MTId, td );
		}

		internal void RegisterThread( string name )
		{
			var td = new ThreadData( name );
			lock ( key ) _ThreadDatas.Add( td.MTId, td );
		}

		float X0 = 0;
		float X1 = 0;
		float X2 = 0;

		// callers must hold lock on key
		void MeasureColumns( Graphics g )
		{
			X0 = 0;
			X1 = X0 + 10 + g.MeasureString( "00.000", _Font ).Width;
			X2 = X1 + 10 + g.MeasureString( "( 00 )", _Font ).Width;

			if ( _ThreadDatas.Count == 0 ) return;

			var indexes = _ThreadDatas.Values.Select( td => td.Index ).Distinct().ToList();
			var min = indexes.Min();
			var max = indexes.Max() + 1;

			float DX = _Size.Width - X1 - 10;
			DX = ( ( DX <= 0 || max <= 0 ) ? 0 : DX / max );

			foreach ( var td in _ThreadDatas.Values )
			{
				td.ThreadX = X1;

				if ( td.Index < 0 )
					td.TextX = X2;
				else
					td.TextX = X1 + 10 + ( td.Index * DX );
			}
		}

		//-----------------------------------------------------------------------------------------
		// Entries

		class EntryData
		{
			public DateTime Timestamp { get; protected set; }
			public int MTId { get; protected set; }
			public string Text { get; set; }

			public EntryData( string text )
			{
				Timestamp = DateTime.Now;
				MTId = Thread.CurrentThread.ManagedThreadId;
				Text = text;
			}
		}

		List<EntryData> _EntryList = new List<EntryData>();

		internal void Add( string text )
		{
			lock ( key ) _EntryList.Add( new EntryData( text ) );

			ConcurrentPaint();

			SetVerticalScroll();
		}

		//-----------------------------------------------------------------------------------------
		// Painting

		protected override void OnPaint( PaintEventArgs e )
		{
			base.OnPaint( e );

			DoPaint( e.Graphics );
		}

		void ConcurrentPaint()
		{
			using ( var g = CreateGraphics() ) DoPaint( g );
		}

		void DoPaint( Graphics g )
		{
			g.Clear( SystemColors.Window );

			if ( _EntryList.Count == 0 ) return;

			g.TranslateTransform( 0, -_VerticalScrollPosition );

			lock ( key )
			{
				MeasureColumns( g );

				float y = DoPaintHeadings( g );
				float dy = 1.5f * g.MeasureString( "M", _Font ).Height;

				var dt = _EntryList[ 0 ].Timestamp;

				foreach ( var entry in _EntryList )
				{
					var time = ( entry.Timestamp - dt ).TotalMilliseconds / 1000;
					DrawString( g, time.ToString( "00.000" ), X0, y );

					ThreadData td = null;
					if ( _ThreadDatas.TryGetValue( entry.MTId, out td ) )
					{
						DrawString( g, entry.Text, td.TextX, y );
					}
					else
					{
						DrawString( g, "( " + entry.MTId.ToString( "00" ) + " )", X1, y );
						DrawString( g, entry.Text, X2, y );
					}

					y += dy;
				}

				_ViewportHeight = ( int ) y;
			}
		}

		private float DoPaintHeadings( Graphics g )
		{
			if ( !_ThreadDatas.Values.Any( td => td.Index >= 0 ) ) return 0;

			var font = new Font( _Font, FontStyle.Bold );

			float dy = 1.2f * g.MeasureString( "m", font ).Height;

			foreach ( var td in _ThreadDatas.Values )
				if ( td.Index >= 0 )
					DrawString( g, font, td.Name, td.TextX, 0 );

			g.DrawLine( Pens.Black, 0, dy, _Size.Width, dy );

			return dy + 10;
		}

		void DrawString( Graphics g, string text, float x, float y )
		{
			DrawString( g, _Font, text, x, y );
		}

		void DrawString( Graphics g, Font font, string text, float x, float y )
		{
			g.DrawString(
				text,
				font,
				SystemBrushes.WindowText,
				x, y );
		}
	}
}
